package de.ozgcloud.eingang.xta.zip;

import java.io.File;
import java.io.IOException;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.springframework.stereotype.Component;
import org.springframework.util.MimeTypeUtils;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.IncomingFile;

// TODO Resolve code duplication with ZipAttachmentReader in semantik-adapter common.zip
// Note: In contrast to the ZipAttachmentReader, here, the zip file is not included in list of extracted files
// Further, there is no ZIP_MAX_THRESHOLD detection.
@Component
public class ZipFileExtractor {

	static final int ZIP_MAX_TOTAL_SIZE = 500 * 1024 * 1024;
	static final int ZIP_MAX_ENTRIES = 100;

	public List<IncomingFile> extractIncomingFilesSafely(IncomingFile zipIncomingFile) {
		var zipFile = zipIncomingFile.getFile();
		verifySizeLimit(zipFile);
		return extractIncomingFiles(zipFile);
	}

	void verifySizeLimit(File zipFile) {
		var totalSize = sumUncompressedEntrySizes(zipFile);
		if (totalSize > ZIP_MAX_TOTAL_SIZE) {
			throw new TechnicalException("Expect uncompressed size %s to be smaller than %d!".formatted(totalSize, ZIP_MAX_TOTAL_SIZE));
		}
	}

	Long sumUncompressedEntrySizes(File zipFile) {
		return mapZipEntries(zipFile, ReadableZipEntry::getSize)
				.stream()
				.mapToLong(Long::longValue)
				.sum();
	}

	List<IncomingFile> extractIncomingFiles(File zipFile) {
		return mapZipEntries(zipFile, this::mapReadableEntryToIncomingFile);
	}

	private IncomingFile mapReadableEntryToIncomingFile(ReadableZipEntry entry) {
		File file;
		try (var inputStream = entry.getInputStream()) {
			file = TempFileUtils.writeTmpFile(inputStream);
		} catch (IOException e) {
			throw new TechnicalException("Failed reading zip file element %s!".formatted(entry.getName()), e);
		}
		return createIncomingFile(file, entry.zipEntry());
	}

	<T> List<T> mapZipEntries(File zipFile, Function<ReadableZipEntry, T> mappingFunction) {
		try (ZipFile zip = new ZipFile(zipFile)) {
			var index = new AtomicInteger();
			var mappedElements = new ArrayList<T>();
			zip.stream().forEach(element -> {
				if (index.getAndIncrement() >= ZIP_MAX_ENTRIES) {
					throw new TechnicalException("Expect zip files to have at most %d entries!".formatted(ZIP_MAX_ENTRIES));
				}
				mappedElements.add(
						mappingFunction.apply(
								ReadableZipEntry.builder()
										.parentZip(zip)
										.zipEntry(element)
										.build()
						)
				);
			});
			return mappedElements;
		} catch (IOException e) {
			throw new TechnicalException("Failed reading zip file!", e);
		}
	}

	IncomingFile createIncomingFile(File file, ZipEntry zipEntry) {
		return IncomingFile.builder()
				.name(zipEntry.getName())
				.size(zipEntry.getSize())
				.contentType(getContentType(zipEntry.getName()))
				.file(file)
				.build();
	}

	String getContentType(String name) {
		Objects.requireNonNull(name);
		return Objects.requireNonNullElse(URLConnection.guessContentTypeFromName(name), MimeTypeUtils.APPLICATION_OCTET_STREAM_VALUE);
	}
}
