package de.ozgcloud.eingang.xta;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.annotation.Bean;
import org.springframework.test.context.ActiveProfiles;

import de.ozgcloud.eingang.Application;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.router.VorgangRemoteService;
import de.ozgcloud.vorgang.vorgang.GrpcEingang;

@SpringBootTest(classes = {
		Application.class,
		XtaITCase.ActivateXTARunnerConfig.class
})
@ActiveProfiles({ "itcase" })
class XtaITCase {

	@MockBean
	private VorgangRemoteService vorgangRemoteService;

	@MockBean
	private XtaRemoteService xtaRemoteService;

	@Captor
	private ArgumentCaptor<FormData> formDataArgumentCaptor;

	@Captor
	private ArgumentCaptor<GrpcEingang> eingangArgumentCaptor;

	@Captor
	private ArgumentCaptor<Optional<String>> organisationseinheitenIdArgumentCaptor;

	@Autowired
	private XtaRunner runner;

	@DisplayName("run get xta messages")
	@Nested
	class TestRunGetXtaMessages {
		@BeforeEach
		void setup() {
			when(vorgangRemoteService.createVorgang(any(), any(), any())).thenReturn("vorgangId(unused)");
		}

		@DisplayName("should have three representations with pdf")
		@TestZipFileNamesWithPdf
		void shouldHaveThreeRepresentationsWithPdf(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfRepresentations()).isEqualTo(3);
		}

		@DisplayName("should send two representations without pdf")
		@TestZipFileNamesWithoutPdf
		void shouldSendTwoXMLRepresentationsWithoutPdf(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfRepresentations()).isEqualTo(2);
		}

		@DisplayName("should have no attachments")
		@TestZipFileNamesWithoutAnlage
		void shouldHaveNoAttachments(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfAttachments()).isZero();
		}

		@DisplayName("should have one attachment")
		@TestZipFileNamesWithAnlage
		void shouldHaveOneAttachment(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var eingang = captureEingang();
			assertThat(eingang.getNumberOfAttachments()).isEqualTo(1);
		}

		@DisplayName("should have postfach id")
		@TestZipFileNamesWithServiceKonto
		void shouldHavePostfachId(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var serviceKonto = captureEingang()
					.getHeader()
					.getServiceKonto();
			assertThat(serviceKonto).isNotNull();
			var postfachId = serviceKonto
					.getPostfachAddresses(0)
					.getIdentifier()
					.getProperty(0)
					.getValue(0);
			assertThat(postfachId).isEqualTo("4dd01647-b9d9-4775-1b50-08da3d83800a");
		}

		@DisplayName("should have organistationseinheit ID")
		@TestZipFileNames
		void shouldHaveOrganistationseinheitId(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var organisationseinheitenId = captureEingang()
					.getZustaendigeStelle()
					.getOrganisationseinheitenId();
			assertThat(organisationseinheitenId).isEqualTo("9795669");
		}

		@DisplayName("should have vorgang nummer")
		@TestZipFileNames
		void shouldHaveVorgangNummer(String zipFileName) {
			mockNachrichtenBroker(zipFileName);

			runner.runGetXtaMessages();

			var vorgangNummer = captureEingang().getHeader().getVorgangNummer();
			assertThat(vorgangNummer).hasSize(9);
		}

		private GrpcEingang captureEingang() {
			verify(vorgangRemoteService, times(1))
					.createVorgang(
							formDataArgumentCaptor.capture(),
							eingangArgumentCaptor.capture(),
							organisationseinheitenIdArgumentCaptor.capture()
					);

			assertThat(formDataArgumentCaptor.getValue()).isNotNull();
			assertThat(eingangArgumentCaptor.getValue()).isNotNull();
			assertThat(organisationseinheitenIdArgumentCaptor.getValue()).isPresent();

			return eingangArgumentCaptor.getValue();
		}
	}

	private void mockNachrichtenBroker(String zipFileName) {
		when(xtaRemoteService.getMessage(any(XtaMessageId.class))).thenReturn(
				XtaResponseTestFactory.createGetMessageResponse(zipFileName)
		);
		when(xtaRemoteService.getMessagesMetadata())
				.thenReturn(XtaResponseTestFactory.createGetStatusListResponse(List.of(zipFileName)));
		when(xtaRemoteService.getNextMessagesMetadata(any()))
				.thenReturn(XtaResponseTestFactory.createEmptyGetStatusListResponse());
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_without_anlage.zip",
			"brauchtumsfeuer_without_anlage.zip",
			"mantelantrag_without_anlage.zip"
	})
	@interface TestZipFileNamesWithoutAnlage {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_with_anlage.zip",
			"brauchtumsfeuer_with_anlage.zip",
			"mantelantrag_with_anlage.zip"
	})
	@interface TestZipFileNamesWithAnlage {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_without_anlage.zip",
			"mantelantrag_without_anlage.zip",
			"brauchtumsfeuer_without_anlage.zip",
			"dfoerdermittel_with_anlage.zip",
			"mantelantrag_with_anlage.zip",
			"brauchtumsfeuer_with_anlage.zip",
	})
	@interface TestZipFileNamesWithPdf {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"waffenschein.zip"
	})
	@interface TestZipFileNamesWithoutPdf {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_without_anlage.zip",
			"dfoerdermittel_with_anlage.zip",
	})
	@interface TestZipFileNamesWithServiceKonto {
	}

	@Target({ ElementType.METHOD })
	@Retention(RetentionPolicy.RUNTIME)
	@ParameterizedTest
	@ValueSource(strings = {
			"dfoerdermittel_without_anlage.zip",
			"mantelantrag_without_anlage.zip",
			"brauchtumsfeuer_without_anlage.zip",
			"dfoerdermittel_with_anlage.zip",
			"mantelantrag_with_anlage.zip",
			"brauchtumsfeuer_with_anlage.zip",
			"waffenschein.zip"
	})
	@interface TestZipFileNames {
	}

	static class ActivateXTARunnerConfig {
		@Bean
		XtaRunner xtaRunner() {
			return new XtaRunner();
		}
	}
}
